@@ -1,14 +1,14 @@
 /*
 SMS Server Tools
-Copyright (C) 2000-2002 Stefan Frings
+Copyright (C) Stefan Frings
 
 This program is free software unless you got it under another license directly
-from the author. You can redistribute it and/or modify it under the terms of 
+from the author. You can redistribute it and/or modify it under the terms of
 the GNU General Public License as published by the Free Software Foundation.
 Either version 2 of the License, or (at your option) any later version.
 
 http://www.isis.de/members/~s.frings
-mailto:s.frings@mail.isis.de
+mailto:stefan.frings@vodafone.de
 */
 
 #include <sys/types.h>
@@ -19,21 +19,27 @@ mailto:s.frings@mail.isis.de
 #include <string.h>
 #include <unistd.h>
 #include <syslog.h>
+#include <time.h>
+#include <stdarg.h>
 #include "charset.h"
 #include "version.h"
 #include "modeminit.h"
 #include "logging.h"
+#include "alarm.h"
 
 char to[255] = {0};
 char subject[255] = {0};
 char outputfile[255] = {0};
 char logfile[255] = {0};
 int loglevel=9;
+char alarmhandler[256]={0};
+int alarmlevel=LOG_WARNING;
 int cs_convert;
 int keep;
-int is_binary=0;       // 1 if the sms is in binary format 
+int is_binary=0;       // 1 if the sms is in binary format
+int with_udh=0;        // 1 if udh bit is set
 int is_statusreport=0; // 1 if the sms is a status report
-
+int quick=0;	       // 1 if quick mode
 
 int octet2bin(char* octet) /* converts an octet to a 8-Bit value */
 {
@@ -61,12 +67,14 @@ int pdu2ascii(char* pdu, char* ascii) /* converts a PDU-String to Ascii */
   int octetcounter;
   char c;
   char binary[500];
-  
+  int ext_characters=0;
+  int extended=0;
+
   /* First convert all octets to bytes */
   count=octet2bin(pdu);
   for (octetcounter=0; octetcounter<count; octetcounter++)
     binary[octetcounter]=octet2bin(pdu+(octetcounter<<1)+2);
-  
+
   /* Then convert from 8-Bit to 7-Bit encapsulated in 8 bit */
   for (charcounter=0; charcounter<count; charcounter++)
   {
@@ -81,14 +89,30 @@ int pdu2ascii(char* pdu, char* ascii) /* converts a PDU-String to Ascii */
       c=(c>>1)&127; /* The shift fills with 1, but I want 0 */
     }
     if (cs_convert)
-      ascii[charcounter]=sms2ascii(c);
+    {
+      if (extended)
+      {
+        ascii[charcounter-ext_characters]=ext_convert(c,CS_SMS,CS_ISO);
+        extended=0;
+      }
+      else
+      {
+        ascii[charcounter-ext_characters]=convert(c,CS_SMS,CS_ISO);
+        // If this is an ESC character, then an extended character code follows
+        if (ascii[charcounter-ext_characters]==0x1B)
+        {
+          ext_characters++;
+          extended=1;
+        }
+      }
+    }
     else if (c==0)
-      ascii[charcounter]=183;
+      ascii[charcounter-ext_characters]=183;
     else
-      ascii[charcounter]=c;
+      ascii[charcounter-ext_characters]=c;
   }
-  ascii[count]=0;
-  return count;
+  ascii[charcounter-ext_characters]=0;
+  return charcounter-ext_characters;
 }
 
 int pdu2binary(char* pdu, char* binary)
@@ -105,15 +129,15 @@ int pdu2binary(char* pdu, char* binary)
 
 int getsms(int sim,char* pdu)  /* reads a SMS from the SIM-memory 1-10 */
 {                              /* returns number of SIM memory if successful */
-			       /* on digicom the return value can be != sim */	
-  char command[5000];
-  char answer[500];
+			       /* on digicom the return value can be != sim */
+  char command[500];
+  char answer[1024];
   char* position;
   char* beginning;
   char* end;
   char tmp[32];
   int i;
-  
+
   if (strcmp(mode,"digicom")==0) // Digicom reports date+time only with AT+CMGL
   {
     writelogfile(LOG_INFO,"Trying to get stored message");
@@ -123,7 +147,7 @@ int getsms(int sim,char* pdu)  /* reads a SMS from the SIM-memory 1-10 */
     for (i=1; i<11; i++)
     {
       sprintf(tmp,"+CMGL: %i",i);
-      position=strstr(answer,tmp); 
+      position=strstr(answer,tmp);
       if (position)
       {
         writelogfile(LOG_INFO,"Found a message at memory %i",i);
@@ -132,7 +156,7 @@ int getsms(int sim,char* pdu)  /* reads a SMS from the SIM-memory 1-10 */
       }
     }
   }
-  
+
   else
   {
     writelogfile(LOG_INFO,"Trying to get stored message %i",sim);
@@ -141,11 +165,12 @@ int getsms(int sim,char* pdu)  /* reads a SMS from the SIM-memory 1-10 */
     /* search for beginning of the answer */
     position=strstr(answer,"+CMGR:");
   }
-  
-  if (position==0) /* keine SMS empfangen, weil Modem nicht mit +CMGR oder +CMGL geantwortet hat */
+
+  if (position==0) /* no SMS received, because modem did not answer with +CMGR or +CMGL  */
     return 0;
+
   beginning=position+7;
-  if (strstr(answer,",,0\r")) /* keine SMS, weil Modem mit +CMGR: 0,,0 geantwortet hat */
+  if (strstr(answer,",,0\r")) /* No SMS,  because Modem answered with +CMGR: 0,,0  */
     return 0;
 
   /* After that we have the PDU or ASCII string */
@@ -163,7 +188,7 @@ int getsms(int sim,char* pdu)  /* reads a SMS from the SIM-memory 1-10 */
 
 void deletesms(int sim) /* deletes the selected sms from the sim card */
 {
-  char command[100];               
+  char command[100];
   char answer[500];
   writelogfile(LOG_INFO,"Deleting message %i",sim);
   sprintf(command,"AT+CMGD=%i\r",sim);
@@ -248,10 +273,10 @@ int splitascii(char* source, char* sendr, char* name, char* date, char* time, ch
   }
   /* Get the date */
   start=end+3;
-  sprintf(date,"%c%c-%c%c-%c%c",start[3],start[4],start[0],start[1],start[6],start[7]);
+  sprintf(date,"%c%c-%c%c-%c%c",start[0],start[1],start[3],start[4],start[6],start[7]);
   /* Get the time */
   start+=9;
-  sprintf(time,"%c%c:%c%c:%c%c",start[0],start[1],start[3],start[4],start[7],start[7]);
+  sprintf(time,"%c%c:%c%c:%c%c",start[0],start[1],start[3],start[4],start[6],start[7]);
   return strlen(ascii);
 }
 
@@ -262,20 +287,34 @@ int split_type_0(char* Pointer,char* sendr,char* date,char* time,char* ascii)
 {
   int Length;
   int padding;
+  char tmpsender[100];
   Length=octet2bin(Pointer);
   padding=Length%2;
-  Pointer+=4;
-  strncpy(sendr,Pointer,Length+padding);
-  swapchars(sendr);
-  /* remove Padding characters after swapping */
-  sendr[Length]=0;
+  Pointer+=2;
+  if ((octet2bin(Pointer) & 112)==80)
+  {  // Sender is alphanumeric
+    Pointer+=2;
+    if (Length>sizeof(tmpsender)-3-padding)  //Limit length of senders name
+      Length=sizeof(tmpsender)-3-padding;
+    snprintf(tmpsender,Length+3+padding,"%02x%s",Length*4/7,Pointer);
+    pdu2ascii(tmpsender,sendr);
+  }
+  else
+  {  // sender is numeric
+    Pointer+=2;
+    strncpy(sendr,Pointer,Length+padding);
+    swapchars(sendr);
+    sendr[Length]=0;
+    if (sendr[Length-1]=='F')
+      sendr[Length-1]=0;
+  }
   Pointer=Pointer+Length+padding+3;
   if ((Pointer[0] & 4)==4)
     is_binary=1;
   else
     is_binary=0;
   Pointer++;
-  sprintf(date,"%c%c-%c%c-%c%c",Pointer[3],Pointer[2],Pointer[5],Pointer[4],Pointer[1],Pointer[0]);
+  sprintf(date,"%c%c-%c%c-%c%c",Pointer[1],Pointer[0],Pointer[3],Pointer[2],Pointer[5],Pointer[4]);
   Pointer=Pointer+6;
   sprintf(time,"%c%c:%c%c:%c%c",Pointer[1],Pointer[0],Pointer[3],Pointer[2],Pointer[5],Pointer[4]);
   Pointer=Pointer+8;
@@ -286,7 +325,7 @@ int split_type_0(char* Pointer,char* sendr,char* date,char* time,char* ascii)
 }
 
 
-// Subroutine for splitpdu() for messages type 2 (Staus Report)
+// Subroutine for splitpdu() for messages type 2 (Status Report)
 // Returns the length of the ascii string. In binary mode ascii contains the binary SMS
 int split_type_2(char* position,char* from, char* date,char* time,char* result)
 {
@@ -294,26 +333,44 @@ int split_type_2(char* position,char* from, char* date,char* time,char* result)
   int padding;
   int status;
   char temp[32];
+  char tmpfrom[100];
+  int messageid;
   strcat(result,"SMS STATUS REPORT\n");
+  // get message id
+  messageid=octet2bin(position);
+  sprintf(temp,"Message_id: %i\n",messageid);
+  strcat(result,temp);
   // get recipient address
   position+=2;
   length=octet2bin(position);
   padding=length%2;
-  position+=4;
-  strncpy(from,position,length+padding);
-  from[length]=0;
-  swapchars(from);
-  strcat(result,"\nDischarge_timestamp: ");
+  position+=2;
+  if ((octet2bin(position) & 112)==80)
+  {  // Sender is alphanumeric
+    position+=2;
+    sprintf(tmpfrom,"%02x%s",length*4/7,position);
+    pdu2ascii(tmpfrom,from);
+  }
+  else
+  {  // sender is numeric
+    position+=2;
+    strncpy(from,position,length+padding);
+    swapchars(from);
+    from[length]=0;
+    if (from[length-1]=='F')
+      from[length-1]=0;
+  }
   // get SMSC timestamp
   position+=length+padding;
-  sprintf(date,"%c%c-%c%c-%c%c",position[3],position[2],position[5],position[4],position[1],position[0]);
+  sprintf(date,"%c%c-%c%c-%c%c",position[1],position[0],position[3],position[2],position[5],position[4]);
   sprintf(time,"%c%c:%c%c:%c%c",position[7],position[6],position[9],position[8],position[11],position[10]);
   // get Discharge timestamp
+  strcat(result,"Discharge_timestamp: ");
   position+=14;
-  sprintf(temp,"%c%c-%c%c-%c%c %c%c:%c%c:%c%c",position[3],position[2],position[5],position[4],position[1],position[0],position[7],position[6],position[9],position[8],position[11],position[10]);
+  sprintf(temp,"%c%c-%c%c-%c%c %c%c:%c%c:%c%c",position[1],position[0],position[3],position[2],position[5],position[4],position[7],position[6],position[9],position[8],position[11],position[10]);
   strcat(result,temp);
-  strcat(result,"\nStatus: ");
   // get Status
+  strcat(result,"\nStatus: ");
   position+=14;
   status=octet2bin(position);
   sprintf(temp,"%i,",status);
@@ -403,14 +460,16 @@ int splitpdu(char* pdu, char* sendr, char* from, char* name, char* date, char* t
       Pointer=pdu+4;
       strncpy(smsc,Pointer,Length);
       swapchars(smsc);
-      /* remove Padding characters after swapping */
+      smsc[Length]=0;
       if (smsc[Length-1]=='F')
         smsc[Length-1]=0;
-      else
-        smsc[Length]=0;
     }
     Pointer=pdu+Length+4;
   }
+  if (octet2bin(Pointer)&4) // Is UDH bit set?
+    with_udh=1;
+  else
+    with_udh=0;
   Type=octet2bin(Pointer) & 3;
   Pointer+=2;
   if (Type==0) // SMS Deliver
@@ -431,20 +490,25 @@ void help()
   printf("Usage:\n");
   printf("              getsms [options]\n");
   printf("Options:\n");
+  printf("              -ax  set alarm handler to x (default none)\n");
+  printf("              -Ax  set alarm level to x (default %i)\n",alarmlevel);
   printf("              -bx  set baudrate to x (default %i)\n",baudrate);
   printf("              -c   use character set conversion\n");
   printf("              -dx  set modem device to x (default %s)\n",device);
   printf("              -ex  wait x seconds before retry (default %d)\n",errorsleeptime);
   printf("              -h   this help\n");
+  printf("              -H   disable hardware handshake RTS/CTS\n");
   printf("              -ix  modem init string x\n");
-  printf("              -k   keep all SMS in the modems memory\n",device);
+  printf("              -k   keep all SMS in the SIM memory\n",device);
   printf("              -lx  use logfile x (filename or handle) (default syslog)\n");
   printf("              -Lx  use loglevel x (default %i)\n",loglevel);
   printf("              -mx  set PDU mode to x (default %s)\n",mode);
   printf("                   x can be old, new or ascii\n");
   printf("              -nx  set modem name to x\n");
-  printf("              -ox  file to write result (default %s)\n",outputfile);
+  printf("              -ox  file result to write (default stdout)\n");
   printf("              -px  set pin to x (only needed if modem not initialized)\n");
+  printf("	        -q   quick initialization\n");
+  printf("              -Q   quicker (check only the first SIM memory)\n");
   printf("              -sx  generate a Subject: line with value x\n");
   printf("              -tx  generate a To: line with value x\n");
   printf("              -V   print version info and copyright\n\n");
@@ -458,7 +522,7 @@ void help()
   printf("              2   cannot verify PIN or modem not ready.\n");
   printf("              3   modem not registered to the network\n");
   printf("              4   no SMS available\n");
-  printf("		5   received SMS Status Report\n");
+  printf("              5   received SMS Status Report\n");
   exit(0);
 }
 
@@ -477,18 +541,29 @@ void parsearguments(int argc,char** argv)
   initstring[0]=0;
   errorsleeptime=10;
   modemname[0]=0;
-
+  rtscts=1;
+  quickinit=0;
   do
   {
-    result=getopt(argc,argv,"n:b:e:o:l:L:khi:cd:t:s:m:p:V");
+    result=getopt(argc,argv,"a:A:n:b:e:o:l:L:khHqQi:cd:t:s:m:p:V");
     switch (result)
     {
-      case 'h': help(); 
+      case 'h': help();
                 break;
+      case 'a': strcpy(alarmhandler,optarg);
+      		break;
+      case 'A': alarmlevel=atoi(optarg);
+      		break;
       case 'k': keep=1;
                 break;
+      case 'H': rtscts=0;
+                break;
       case 'c': cs_convert=1;
       		break;
+      case 'q': quickinit=1;
+                break;
+      case 'Q': quick=1;
+      		break;
       case 'd': strcpy(device,optarg);
                 break;
       case 'e': errorsleeptime=atoi(optarg);
@@ -501,7 +576,7 @@ void parsearguments(int argc,char** argv)
       case 'l': strcpy(logfile,optarg);
                 break;
       case 'L': loglevel=atoi(optarg);
-                break;		
+                break;
       case 'm': strcpy(mode,optarg);
                 break;
       case 'n': strcpy(modemname,optarg);
@@ -514,15 +589,18 @@ void parsearguments(int argc,char** argv)
                 break;
       case 'o': strcpy(outputfile, optarg);
                 break;
-      case 'V': printf("Version %s, Copyright (c) 2000-2002 by Stefan Frings, s.frings@mail.isis.de\n",getsms_version);
+      case 'V': printf("Version %s, Copyright (c) Stefan Frings, stefan.frings@vodafone.de\n",getsms_version);
                 exit(0);
     }
   }
   while (result>0);
-  
+
   if (modemname[0]==0)
     strcpy(modemname,device);
-  
+}
+
+void check_arguments(int argc,char** argv)
+{
   switch (baudrate)
   {
     case 300:   baudrate=B300; break;
@@ -549,24 +627,33 @@ int main(int argc,char** argv)
 {
   int userdatalength;
   int sim;
-  char pdu[500];
-  char ascii[500];
-  char sendr[31];
+  char pdu[1024];
+  char ascii[1024];
+  char sendr[100];
   char smsc[31];
   char name[64];
   char date[9];
-  char time[9];
+  char Time[9];
   int found;
   char tmp[100];
   int max_memory,used_memory;
+  FILE *fd;
   parsearguments(argc,argv);
   snprintf(tmp,sizeof(tmp),"getsms (%s)",modemname); tmp[sizeof(tmp)-1]=0;
   openlogfile(tmp,logfile,LOG_DAEMON,loglevel);
+  set_alarmhandler(alarmhandler,alarmlevel,modemname);
+  check_arguments(argc,argv);
   writelogfile(LOG_INFO,"Checking incoming SMS");
   openmodem();
   setmodemparams();
   initmodem();
-  check_memory(&used_memory,&max_memory);
+  if (!quick)
+    check_memory(&used_memory,&max_memory);
+  else
+  {
+    used_memory=1;
+    max_memory=1;
+  }
   found=0;
   if (used_memory)
   {
@@ -576,63 +663,64 @@ int main(int argc,char** argv)
       {
         // Ok, now we split the PDU string into parts and show it
         if ((strcmp(mode,"ascii")==0) || (strcmp(mode,"digicom")==0))
-          userdatalength=splitascii(pdu,sendr,name,date,time,ascii);
+          userdatalength=splitascii(pdu,sendr,name,date,Time,ascii);
         else
-	  userdatalength=splitpdu(pdu,sendr,to,name,date,time,ascii,smsc);
-        if (outputfile[0]==0)
-        {
-          printf("From: %s\n",sendr);
-	  if (smsc[0])
-	    printf("From_SMSC: %s\n",smsc);
-	  if (name[0])
-	    printf("Name: %s\n",name);
-          if (to[0])
-            printf("To: %s\n",to);
-	  if (date[0] || time[0])
-            printf("Sent: %s %s\n",date,time);
-          if (subject[0])
-            printf("Subject: %s\n\n",subject);
- 	  if (is_binary==0)
-            printf("%s\n\n",ascii);
-	  else
-	    printf("Message has %i bytes of binary data\n\n",userdatalength);
-        }
+	  userdatalength=splitpdu(pdu,sendr,to,name,date,Time,ascii,smsc);
 
-        writelogfile(LOG_NOTICE, "SMS received, From: %s, Sent: %s %s",sendr,date,time);
+	writelogfile(LOG_NOTICE, "SMS received, From: %s",sendr);
+	alarm_handler(LOG_NOTICE,"", "SMS received, From: %s",sendr);
         if (is_binary==0)
-          writelogfile(LOG_INFO,"Message: %s", ascii);
+          writelogfile(LOG_DEBUG,"Message: %s", ascii);
         else
-          writelogfile(LOG_INFO,"Message has binary data\n\n",userdatalength);
+          writelogfile(LOG_DEBUG,"Message has binary data\n\n",userdatalength);
 
         if (outputfile[0])
-        {
-	  FILE *fd;
 	  fd = fopen(outputfile, "w");
-	  if (fd)
+	else
+	  fd=stdout;
+	if (fd)
+	{
+	  fprintf(fd, "From: %s\n",sendr);
+	  if (smsc[0])
+	    fprintf(fd,"From_SMSC: %s\n",smsc);
+	  if (name[0])
+	    fprintf(fd,"Name: %s\n",name);
+	  if (to[0])
+	    fprintf(fd, "To: %s\n",to);
+	  if (date[0] || Time[0])
+	    fprintf(fd, "Sent: %s %s\n",date,Time);
+	  // Add local timestamp
 	  {
-	    fprintf(fd, "From: %s\n",sendr);
-	    if (smsc[0])
-	      fprintf(fd,"From_SMSC: %s\n",smsc);
-	    if (name[0])
-	      fprintf(fd,"Name: %s\n",name);
-	    if (to[0])
-	      fprintf(fd, "To: %s\n",to);
-	    if (date[0] || time[0])
-	      fprintf(fd, "Sent: %s %s\n",date,time);
-	    if (subject[0])
-	      fprintf(fd, "Subject: %s\n",subject);
-            if (is_binary==0)
-  	      fprintf(fd, "\n%s\n",ascii);
-	    else
-	    {
-	      fprintf(fd,"Binary: true\n\n");
-	      fwrite(ascii,1,userdatalength,fd);
-	    }
-	    fclose(fd);
-	  }
+	    char timestamp[40];
+            time_t now;
+            time(&now);
+            strftime(timestamp,sizeof(timestamp),"%y-%m-%d %H:%M:%S",localtime(&now));
+	    fprintf(fd,"Received: %s\n",timestamp);
+          }
+	  if (subject[0])
+	    fprintf(fd, "Subject: %s\n",subject);
+          if (is_binary==0)
+  	    fprintf(fd, "\n%s\n",ascii);
 	  else
-	    writelogfile(LOG_ERR,"Cannot create file %s!", outputfile);
-        }
+	  {
+	    if (with_udh)
+	      fprintf(fd,"UDH: true\n");
+	    else
+	      fprintf(fd,"UDH: false\n");
+	    fprintf(fd,"Binary: true\n\n");
+	    if (outputfile[0])
+	      fwrite(ascii,1,userdatalength,fd);
+	    else
+	      fprintf(fd, "binary not displayed\n");
+	  }
+	  if (outputfile[0]) 
+            fclose(fd);
+	}
+	else
+	{
+	  writelogfile(LOG_ERR,"Cannot create file %s!", outputfile);
+	  alarm_handler(LOG_ERR,"","Cannot create file %s!", outputfile);
+	}
         if (keep==0)
           deletesms(found);
         break; // we want only get the first available SMS
@@ -651,5 +739,8 @@ int main(int argc,char** argv)
     writelogfile(LOG_INFO,"No SMS available");
     exit(4);
   }
-  return 0;
+  if (is_statusreport)
+    return 5;
+  else
+    return 0;
 }
