@@ -1,6 +1,6 @@
 /*
 SMS Server Tools
-Copyright (C) 2000 Stefan Frings
+Copyright (C) Stefan Frings
 
 This program is free software unless you got it under another license directly
 from the author. You can redistribute it and/or modify it under the terms of 
@@ -8,7 +8,7 @@ the GNU General Public License as published by the Free Software Foundation.
 Either version 2 of the License, or (at your option) any later version.
 
 http://www.isis.de/members/~s.frings
-mailto:s.frings@mail.isis.de
+mailto:stefan.frings@vodafone.de
 */
 
 #ifndef LOGGING_H
